@@ -15,7 +15,6 @@ mod infer;
 mod interner;
 mod lower;
 mod mapping;
-mod op;
 mod tls;
 mod utils;
 mod walk;
@@ -36,12 +35,12 @@ use std::sync::Arc;
 use chalk_ir::{
     fold::{Fold, Shift},
     interner::HasInterner,
-    UintTy,
+    NoSolution, UintTy,
 };
 use hir_def::{
     expr::ExprId,
     type_ref::{ConstScalar, Rawness},
-    TypeParamId,
+    TypeOrConstParamId,
 };
 
 use crate::{db::HirDatabase, utils::generics};
@@ -56,9 +55,9 @@ pub use lower::{
     TyDefId, TyLoweringContext, ValueTyDefId,
 };
 pub use mapping::{
-    const_from_placeholder_idx, from_assoc_type_id, from_chalk_trait_id, from_foreign_def_id,
-    from_placeholder_idx, lt_from_placeholder_idx, to_assoc_type_id, to_chalk_trait_id,
-    to_foreign_def_id, to_placeholder_idx,
+    from_assoc_type_id, from_chalk_trait_id, from_foreign_def_id, from_placeholder_idx,
+    lt_from_placeholder_idx, to_assoc_type_id, to_chalk_trait_id, to_foreign_def_id,
+    to_placeholder_idx,
 };
 pub use traits::TraitEnvironment;
 pub use utils::all_super_traits;
@@ -112,6 +111,7 @@ pub type Canonical<T> = chalk_ir::Canonical<T>;
 pub type FnSig = chalk_ir::FnSig<Interner>;
 
 pub type InEnvironment<T> = chalk_ir::InEnvironment<T>;
+pub type Environment = chalk_ir::Environment<Interner>;
 pub type DomainGoal = chalk_ir::DomainGoal<Interner>;
 pub type Goal = chalk_ir::Goal<Interner>;
 pub type AliasEq = chalk_ir::AliasEq<Interner>;
@@ -123,13 +123,13 @@ pub type WhereClause = chalk_ir::WhereClause<Interner>;
 // FIXME: get rid of this
 pub fn subst_prefix(s: &Substitution, n: usize) -> Substitution {
     Substitution::from_iter(
-        &Interner,
-        s.as_slice(&Interner)[..std::cmp::min(s.len(&Interner), n)].iter().cloned(),
+        Interner,
+        s.as_slice(Interner)[..std::cmp::min(s.len(Interner), n)].iter().cloned(),
     )
 }
 
 /// Return an index of a parameter in the generic type parameter list by it's id.
-pub fn param_idx(db: &dyn HirDatabase, id: TypeParamId) -> Option<usize> {
+pub fn param_idx(db: &dyn HirDatabase, id: TypeOrConstParamId) -> Option<usize> {
     generics(db.upcast(), id.parent).param_idx(id)
 }
 
@@ -137,7 +137,7 @@ pub(crate) fn wrap_empty_binders<T>(value: T) -> Binders<T>
 where
     T: Fold<Interner, Result = T> + HasInterner<Interner = Interner>,
 {
-    Binders::empty(&Interner, value.shifted_in_from(&Interner, DebruijnIndex::ONE))
+    Binders::empty(Interner, value.shifted_in_from(Interner, DebruijnIndex::ONE))
 }
 
 pub(crate) fn make_only_type_binders<T: HasInterner<Interner = Interner>>(
@@ -146,7 +146,7 @@ pub(crate) fn make_only_type_binders<T: HasInterner<Interner = Interner>>(
 ) -> Binders<T> {
     Binders::new(
         VariableKinds::from_iter(
-            &Interner,
+            Interner,
             std::iter::repeat(chalk_ir::VariableKind::Ty(chalk_ir::TyVariableKind::General))
                 .take(num_vars),
         ),
@@ -165,7 +165,7 @@ pub fn make_canonical<T: HasInterner<Interner = Interner>>(
             chalk_ir::UniverseIndex::ROOT,
         )
     });
-    Canonical { value, binders: chalk_ir::CanonicalVarKinds::from_iter(&Interner, kinds) }
+    Canonical { value, binders: chalk_ir::CanonicalVarKinds::from_iter(Interner, kinds) }
 }
 
 // FIXME: get rid of this, just replace it by FnPointer
@@ -175,6 +175,7 @@ pub fn make_canonical<T: HasInterner<Interner = Interner>>(
 pub struct CallableSig {
     params_and_return: Arc<[Ty]>,
     is_varargs: bool,
+    legacy_const_generics_indices: Arc<[u32]>,
 }
 
 has_interner!(CallableSig);
@@ -185,7 +186,11 @@ pub type PolyFnSig = Binders<CallableSig>;
 impl CallableSig {
     pub fn from_params_and_return(mut params: Vec<Ty>, ret: Ty, is_varargs: bool) -> CallableSig {
         params.push(ret);
-        CallableSig { params_and_return: params.into(), is_varargs }
+        CallableSig {
+            params_and_return: params.into(),
+            is_varargs,
+            legacy_const_generics_indices: Arc::new([]),
+        }
     }
 
     pub fn from_fn_ptr(fn_ptr: &FnPointer) -> CallableSig {
@@ -194,23 +199,28 @@ impl CallableSig {
             params_and_return: fn_ptr
                 .substitution
                 .clone()
-                .shifted_out_to(&Interner, DebruijnIndex::ONE)
+                .shifted_out_to(Interner, DebruijnIndex::ONE)
                 .expect("unexpected lifetime vars in fn ptr")
                 .0
-                .as_slice(&Interner)
+                .as_slice(Interner)
                 .iter()
-                .map(|arg| arg.assert_ty_ref(&Interner).clone())
+                .map(|arg| arg.assert_ty_ref(Interner).clone())
                 .collect(),
             is_varargs: fn_ptr.sig.variadic,
+            legacy_const_generics_indices: Arc::new([]),
         }
     }
 
+    pub fn set_legacy_const_generics_indices(&mut self, indices: &[u32]) {
+        self.legacy_const_generics_indices = indices.into();
+    }
+
     pub fn to_fn_ptr(&self) -> FnPointer {
         FnPointer {
             num_binders: 0,
             sig: FnSig { abi: (), safety: Safety::Safe, variadic: self.is_varargs },
             substitution: FnSubst(Substitution::from_iter(
-                &Interner,
+                Interner,
                 self.params_and_return.iter().cloned(),
             )),
         }
@@ -228,17 +238,18 @@ impl CallableSig {
 impl Fold<Interner> for CallableSig {
     type Result = CallableSig;
 
-    fn fold_with<'i>(
+    fn fold_with<E>(
         self,
-        folder: &mut dyn chalk_ir::fold::Folder<'i, Interner>,
+        folder: &mut dyn chalk_ir::fold::Folder<Interner, Error = E>,
         outer_binder: DebruijnIndex,
-    ) -> chalk_ir::Fallible<Self::Result>
-    where
-        Interner: 'i,
-    {
+    ) -> Result<Self::Result, E> {
         let vec = self.params_and_return.to_vec();
         let folded = vec.fold_with(folder, outer_binder)?;
-        Ok(CallableSig { params_and_return: folded.into(), is_varargs: self.is_varargs })
+        Ok(CallableSig {
+            params_and_return: folded.into(),
+            is_varargs: self.is_varargs,
+            legacy_const_generics_indices: self.legacy_const_generics_indices,
+        })
     }
 }
 
@@ -261,18 +272,18 @@ pub(crate) struct ReturnTypeImplTrait {
 }
 
 pub fn static_lifetime() -> Lifetime {
-    LifetimeData::Static.intern(&Interner)
+    LifetimeData::Static.intern(Interner)
 }
 
 pub fn dummy_usize_const() -> Const {
-    let usize_ty = chalk_ir::TyKind::Scalar(Scalar::Uint(UintTy::Usize)).intern(&Interner);
+    let usize_ty = chalk_ir::TyKind::Scalar(Scalar::Uint(UintTy::Usize)).intern(Interner);
     chalk_ir::ConstData {
         ty: usize_ty,
         value: chalk_ir::ConstValue::Concrete(chalk_ir::ConcreteConst {
             interned: ConstScalar::Unknown,
         }),
     }
-    .intern(&Interner)
+    .intern(Interner)
 }
 
 pub(crate) fn fold_free_vars<T: HasInterner<Interner = Interner> + Fold<Interner>>(
@@ -281,13 +292,15 @@ pub(crate) fn fold_free_vars<T: HasInterner<Interner = Interner> + Fold<Interner
 ) -> T::Result {
     use chalk_ir::{fold::Folder, Fallible};
     struct FreeVarFolder<F>(F);
-    impl<'i, F: FnMut(BoundVar, DebruijnIndex) -> Ty + 'i> Folder<'i, Interner> for FreeVarFolder<F> {
-        fn as_dyn(&mut self) -> &mut dyn Folder<'i, Interner> {
+    impl<'i, F: FnMut(BoundVar, DebruijnIndex) -> Ty + 'i> Folder<Interner> for FreeVarFolder<F> {
+        type Error = NoSolution;
+
+        fn as_dyn(&mut self) -> &mut dyn Folder<Interner, Error = Self::Error> {
             self
         }
 
-        fn interner(&self) -> &'i Interner {
-            &Interner
+        fn interner(&self) -> Interner {
+            Interner
         }
 
         fn fold_free_var_ty(
@@ -311,13 +324,15 @@ pub(crate) fn fold_tys<T: HasInterner<Interner = Interner> + Fold<Interner>>(
         Fallible,
     };
     struct TyFolder<F>(F);
-    impl<'i, F: FnMut(Ty, DebruijnIndex) -> Ty + 'i> Folder<'i, Interner> for TyFolder<F> {
-        fn as_dyn(&mut self) -> &mut dyn Folder<'i, Interner> {
+    impl<'i, F: FnMut(Ty, DebruijnIndex) -> Ty + 'i> Folder<Interner> for TyFolder<F> {
+        type Error = NoSolution;
+
+        fn as_dyn(&mut self) -> &mut dyn Folder<Interner, Error = Self::Error> {
             self
         }
 
-        fn interner(&self) -> &'i Interner {
-            &Interner
+        fn interner(&self) -> Interner {
+            Interner
         }
 
         fn fold_ty(&mut self, ty: Ty, outer_binder: DebruijnIndex) -> Fallible<Ty> {
@@ -338,25 +353,27 @@ where
 {
     use chalk_ir::{
         fold::{Folder, SuperFold},
-        Fallible, NoSolution,
+        Fallible,
     };
     struct ErrorReplacer {
         vars: usize,
     }
-    impl<'i> Folder<'i, Interner> for ErrorReplacer {
-        fn as_dyn(&mut self) -> &mut dyn Folder<'i, Interner> {
+    impl Folder<Interner> for ErrorReplacer {
+        type Error = NoSolution;
+
+        fn as_dyn(&mut self) -> &mut dyn Folder<Interner, Error = Self::Error> {
             self
         }
 
-        fn interner(&self) -> &'i Interner {
-            &Interner
+        fn interner(&self) -> Interner {
+            Interner
         }
 
         fn fold_ty(&mut self, ty: Ty, outer_binder: DebruijnIndex) -> Fallible<Ty> {
-            if let TyKind::Error = ty.kind(&Interner) {
+            if let TyKind::Error = ty.kind(Interner) {
                 let index = self.vars;
                 self.vars += 1;
-                Ok(TyKind::BoundVar(BoundVar::new(outer_binder, index)).intern(&Interner))
+                Ok(TyKind::BoundVar(BoundVar::new(outer_binder, index)).intern(Interner))
             } else {
                 let ty = ty.super_fold_with(self.as_dyn(), outer_binder)?;
                 Ok(ty)
@@ -374,7 +391,7 @@ where
                 // won't contain the whole thing, which would not be very helpful
                 Err(NoSolution)
             } else {
-                Ok(TyKind::Error.intern(&Interner))
+                Ok(TyKind::Error.intern(Interner))
             }
         }
 
@@ -388,7 +405,7 @@ where
                 // won't contain the whole thing, which would not be very helpful
                 Err(NoSolution)
             } else {
-                Ok(TyKind::Error.intern(&Interner))
+                Ok(TyKind::Error.intern(Interner))
             }
         }
 
@@ -453,5 +470,5 @@ where
             chalk_ir::UniverseIndex::ROOT,
         )
     });
-    Canonical { value, binders: chalk_ir::CanonicalVarKinds::from_iter(&Interner, kinds) }
+    Canonical { value, binders: chalk_ir::CanonicalVarKinds::from_iter(Interner, kinds) }
 }
